import base64
import hashlib
import hmac
from io import BytesIO
import json
import os
import requests
import time
import urllib.parse
import werkzeug.formparser

from cryptography.fernet import Fernet

try:
    from const import TIMEOUT, SIGNED_PARAMS_SHARED_SECRET
except ImportError:
    from .const import TIMEOUT, SIGNED_PARAMS_SHARED_SECRET

# This file contains the functions that are specific to the SaasAuth authentication method.
# They were split from the original util.py file as util is meant to be run inside an odoo server (lots of odoo import)
# while we would like to be able to call SaasAuth for standalone scripts.
# the functions in this file are imported inside util.py to maintain the current behaviour.

def _sign_message(message):
    """Sign the message with the http_shared_secret from the current server

        :param str message: The message should be crafted from the request the following way:
            "url path|query params|body content" and
            if the request does not have any body "url path|query params"

        The timestamp is added to the message before the signature

        :return dict:
            The current server name
            The signature of the message
            The timestamp of in the signature

    """
    me = SaasAuth._get_saas_config().get('me', {})
    if not me:
        raise Exception('Unable to find own server configuration for request signature')

    server_http_secret = me['http_shared_secret']
    h = hmac.new(base64.b64decode(server_http_secret), digestmod=hashlib.sha256)
    msg_timestamp = int(time.time())

    message = '%s|%s' % (msg_timestamp, message)
    h.update(message.encode('utf-8'))  # digest the message
    return {
        'oe-server-name': me['name'],
        'oe-signature': base64.b64encode(h.digest()),
        'oe-timestamp': msg_timestamp,
    }


def sign_params(scope, params, sig_param="_ms"):
    """Return the given `params` dict, modified to included the
    sig_param parameter, depending on the current server
    and time. The result is meant to be used to construct a
    signed URL, and will include the following extra parameters:
     - oe-server-name
     - oe-timestamp
     - oe-signature

    Those parameter are hidden in the `sig_param` parameter
    and symmetrically encrypted.

    The URL that will be called with this parameters should
    be a route that is decorated with `from_role(mode="params", sig_param=`sig_param`)`,
    which will handle the signature parameter.

    :param dict params: the route parameters to sign
    :param str scope: the route that will be called
    :param sig_param: name of the parameter where the signature will be stored
    """
    message = '%s|%s' % (
        scope,
        json.dumps(params, sort_keys=True),  # url query params sorted by key
    )
    signature = _sign_message(message)
    signature_rep = '%s|%s|%s' % (
        signature['oe-server-name'],
        signature['oe-signature'].decode(),
        signature['oe-timestamp'],
    )
    return dict(params, **{sig_param: tokenize_param(signature_rep)})

class SaasAuth(requests.auth.AuthBase):
    """ Use this class as the auth parameter to sign your http request using the `requests` lib

        e.g.: Trial server contacting accounts with a signed http request

            Trial side:

                requests.get(WEBSITE + '/odoo-enterprise/example', auth=SaasAuth())

            Accounts side:

                @from_role('trial')
                @route('/odoo-enterprise/example', type='http', methods=['GET'])
                    return 'Request signature worked perfectly :)'


            By using `auth=SaasAuth()` when doing a http request, the request will be signed and
            the signature is added on the request headers.
            On the reception side, `from_role` has been set to verifiy the integrity of the request.
            If the signature doesn't match, then Forbidden is raised, otherwise the controller is called :)

    """

    @staticmethod
    def _get_saas_config():
        try:
            with open(os.path.expanduser('~odoo/config/config.json')) as c:
                saas_config = json.load(c)
        except (IOError, ValueError):
            saas_config = {}
        return saas_config

    def __call__(self, request):
        parsed_url = urllib.parse.urlparse(request.path_url)
        if request.headers.get('Content-Type') == 'application/json':
            body = request.body
            if isinstance(body, bytes):
                body = body.decode('utf-8')
            body = json.loads(body)
        else:
            if isinstance(request.body, str):
                input_data = request.body.encode()
            else:
                input_data = request.body
            input_stream = BytesIO(input_data)
            environ = {
                'wsgi.input': input_stream,
                'CONTENT_LENGTH': request.headers.get('Content-Length', '0'),
                'CONTENT_TYPE': request.headers.get('Content-Type', 'application/x-www-form-urlencoded'),
                'REQUEST_METHOD': request.method,
            }
            # see https://werkzeug.palletsprojects.com/en/3.0.x/http/
            body = werkzeug.formparser.parse_form_data(environ)[1]

        message = '%s|%s|%s' % (
            parsed_url.path,  # url path
            json.dumps(dict(urllib.parse.parse_qsl(parsed_url.query)), sort_keys=True),  # url query params sorted by key
            json.dumps(body, sort_keys=True)  # http request body
        )

        request.headers.update(_sign_message(message))
        return request


def check_message_signature(message, msg_signature, msg_timestamp, server_http_secret, sig_ttl=TIMEOUT):
    """
    Verify the message matches its signature using the shared secret of the source server.
    Also verify whether the message signature has expired. The signature is valid for sig_ttl
    seconds.

    :param str message: The message to authenticate
    :param bytes msg_signature: The signature of the message
    :param str msg_timestamp: The timestamp of the message
    :param bytes server_http_secret: The shared secret of the source server
    :param int sig_ttl: The time to live of the signature (in seconds). Defaults to TIMEOUT seconds.

    :return bool: True if the signature is valid, False otherwise
    """

    if int(msg_timestamp) + sig_ttl < int(time.time()):
        return False

    h = hmac.new(base64.b64decode(server_http_secret), digestmod=hashlib.sha256)
    h.update(message.encode('utf-8'))
    if not hmac.compare_digest(h.digest(), base64.b64decode(msg_signature)):
        return False

    return True


def extract_signature_from_kwargs(scope, kwargs, sig_param, sig_ttl):
    """
        :param dict query_string: key and values contained in the query string
                                  typically returned by urllib.parse.parse_qs
        :param str sig_param: name of the field that contains the signature
    """
    kwargs = dict(kwargs)
    signature = kwargs.pop(sig_param)
    signature = detokenize_param(signature, token_ttl=sig_ttl)
    msg_server_src, msg_signature, msg_timestamp = signature.split('|')

    message = '%s|%s|%s' % (
        msg_timestamp,
        scope,
        json.dumps(kwargs, sort_keys=True),
    )
    return message, msg_signature, msg_timestamp, msg_server_src

def tokenize_param(param):
    """ Symetrically encrypt the argument with a shared secret common to all servers.

    This is not suitable for long-term encryption, merely for wrapping transient
    tokens or parameters that will be passed in customer-facing URLs, but have
    *no direct impact* to security.
    """
    fernet = Fernet(SIGNED_PARAMS_SHARED_SECRET)
    token = fernet.encrypt(param.encode())
    return token.decode()

def detokenize_param(param, token_ttl=None):
    """ Return the original value of a param protected with tokenize_param
    """
    fernet = Fernet(SIGNED_PARAMS_SHARED_SECRET)
    return fernet.decrypt(param.encode(), ttl=token_ttl).decode()
